module Pico
  class Application < Context
    attr :autoload_paths, :injector

    def initialize(name, root: nil, **params, &config_block)
      super name, root: root, **params
      @autoload_paths = []
      @injector = build_injector
      configure(config_block) if block_given?
    end

    def configure(config_block)
      ConfigurationContext.evaluate(config_block, application: self)
    end

    def boot!
      self.root ||= self.class.default_root
      super
    end

    def build_const_resolver(expanded_const)
      resolver = super
      resolver.autoload_paths = autoload_paths
      resolver
    end

    def possible_implicit_namespace?(path)
      autoload_paths.any? do |autoload_path|
        root.join(autoload_path, path).directory?
      end
    end

    def root=(dir)
      raise ArgumentError, "Cannot alter root after booting" if booted?
      @root = Pathname(dir)
    end

    private

    def build_injector
      injector = Ruse.create_injector
      injector.pico_context = self
      injector
    end

    def build_mod
      ApplicationModule.new self
    end

    class << self
      def default_root
        caller_locations.reduce do |previous_location, location|
          if previous_location.label == "boot!" && File.basename(location.path) != "pico.rb"
            return Pathname(location.absolute_path).dirname
          end
          location
        end
      end
    end

    class ApplicationModule < Module
      attr :application

      def initialize(application)
        @application = application
        extend self
      end

      def build(const_name, **params)
        child_injector = application.injector.clone
        child_injector.configure values: params
        child_injector.get const_name
      rescue Ruse::UnknownServiceError => ruse_error
        raise Exception, "could not resolve dependency `#{ruse_error}'"
      end
    end

    class ConfigurationContext
      def initialize(application)
        define_singleton_method :__application__ do
          application
        end
      end

      def self.evaluate(config_block, application:)
        new(application).instance_eval &config_block
      end

      def autoload_paths
        __application__.autoload_paths
      end

      def autoload_paths=(paths)
        autoload_paths.clear
        autoload_paths.concat paths
      end

      def provide(provision = nil, as:, &factory_block)
        if block_given?
          raise ArgumentError, "cannot supply a block and a value" if provision
          provision = factory_block
        end
        config_key, value = extract_ruse_config provision
        __application__.injector.configure config_key => { as.to_sym => value }
      end

      private

      def extract_ruse_config(provision)
        case provision
        when Class then [:aliases, provision.name]
        when String then [:aliases, provision]
        when -> p { p.respond_to?(:call) } then [:factories, provision]
        else [:values, provision]
        end
      end
    end

  end
end
